#include "synth.h"
#include "synth_math.h"
#include "hardware.h"


#include "xbasic_types.h"
#include "xparameters.h"
#include "xuartns550.h"
#include "xintc.h"
#include "xgpio.h"
#include "mb_interface.h"

static u8 global_volume = 255;
static u8 global_pan = 128;


//This function determines pitch for all operators at the current channels.
//The code here was initially a part of the note_on call, but since the pitch
//bend wheel should be able to modify the pitch of an already pressed key, this
//separation was needed.
void synth_send_pitch(synth_t* synth_handle, u8 channel)
{
	u8 i, note;
	u32 pitch;
	//Even if no note is played, look it up and send it anyway. A decaying release note
	//should be bent by pitch wheel as well.
	note = synth_handle->ch_array[channel] & 0x7F;
	for(i = 0; i < 6; ++i)
	{
		if(synth_handle->patch.osc[i].mode) //fixed mode
			pitch = synth_handle->patch.osc[i].freq_fixed + (synth_handle->patch.osc[i].detune > 7 ? synth_handle->patch.osc[i].detune - 7 : 0);
		else //ratio mode
			pitch = calc_freq(note, (u8)synth_handle->patch.key_transpose, (u8)synth_handle->patch.osc[i].freq_coarse, (u8)synth_handle->patch.osc[i].freq_fine) + synth_handle->patch.osc[i].detune - 7;
		hw_send_op_freq(channel,i,apply_pw(pitch, synth_handle->pitch_bend));
	}
}

//release sustain pedal. Notes set to off should be set to off in HW, since SW doesn't send the note off
//HW message if the sustain pedal is down
void synth_release_sustained(synth_t* synth_handle)
{
	u8 voice;
	for(voice = synth_handle->base_ch; voice < (synth_handle->base_ch + synth_handle->no_ch_avail);++voice)
	{		
		if(!(synth_handle->ch_array[voice] & 0x80))
			hw_send_others(voice,synth_handle->patch.lfo_speed,synth_handle->patch.lfo_sync,synth_handle->patch.lfo_wave,0);
	}
}
// this function sets all hw parameter to their corresponding SW values. SW only
// parameters will not be taken into account and all notes will be set to off
void synth_refresh_hw(synth_t* synth_handle){
	u8 i,j,s;
	u8 curvoice;
	
	for(i = 0; i < synth_handle->no_ch_avail; ++i)
	{
		curvoice = synth_handle->base_ch + i;
		hw_send_algfb(curvoice,synth_handle->patch.alg,synth_handle->patch.fb);
		
		// I think it's pretty obvious that this function also should release 
		// all notes currently played, hence the 0 at the end of the hw_send_others()-call.
		hw_send_others(curvoice,synth_handle->patch.lfo_speed,synth_handle->patch.lfo_sync,synth_handle->patch.lfo_wave,0);
		hw_send_pms(curvoice,synth_handle->patch.pms);
		hw_send_lfo_delay(curvoice,synth_handle->patch.lfo_delay);
		for(s = 0; s < 4; ++s)
		{
			hw_send_peg_rate(curvoice,s,synth_handle->patch.pitch_eg_rate[s]);
			hw_send_peg_lvl(curvoice,s,synth_handle->patch.pitch_eg_lvl[s]);
		}
		// Transpose is a SW only parameter so not sent here,
		// and that goes for the different scaling parameters as well
		for(j = 0; j < 6; ++j)
		{
			hw_send_op_levelpars(curvoice,j,synth_handle->patch.osc[j].output_lvl,synth_handle->patch.osc[j].sync,synth_handle->patch.osc[j].enabled);
			hw_send_op_amodsense(curvoice,j,synth_handle->patch.osc[j].ams);
			for(s = 0; s < 4; ++s)
			{
				hw_send_op_rate(curvoice,j,s,synth_handle->patch.osc[j].eg_rate[s]);
				hw_send_op_eglvl(curvoice,j,s,synth_handle->patch.osc[j].eg_lvl[s]);
			}
			// since frequency depends on note and other SW parameters, 
			// it will not be sent since it's a SW value even though the HW parameter exists
		}
	}
	hw_send_mixer(VOLUME, global_volume);
	hw_send_mixer(PAN, global_pan);
}

//This functions initializes all parameter to a default state.
void synth_init(synth_t* synth_handle, u8 polyphony, u8 base_ch){
	u8 i,j;
	
	//synth parameters
	synth_handle->base_ch = base_ch;
	synth_handle->next_ch = base_ch;
	synth_handle->no_ch_avail = polyphony;
	synth_handle->sustain = 0;
	for(i = 0; i < 16; ++i)
		synth_handle->ch_array[i] = 0;
	synth_handle->pitch_bend = 8192; // wheel center
	synth_handle->patch_no = 0;
	
	//reset current patch
	synth_handle->patch.alg = 0;
	synth_handle->patch.fb = 0;
	synth_handle->patch.lfo_wave = LFO_SQUARE;
	synth_handle->patch.lfo_speed = 15;
	synth_handle->patch.lfo_delay = 0x80000;
	synth_handle->patch.lfo_sync = 0;
	synth_handle->patch.pms = 0;
	for(i = 0; i < 4; ++i)
	{
		synth_handle->patch.pitch_eg_rate[i] = 0x0;
		synth_handle->patch.pitch_eg_lvl[i] = 0x800;
	}
	synth_handle->patch.key_transpose = TRANSPOSE_MID;
	for(i = 0; i < 6; ++i)
	{
		synth_handle->patch.osc[i].ams = 0;
		synth_handle->patch.osc[i].mode = 0;
		synth_handle->patch.osc[i].sync = 0;
		synth_handle->patch.osc[i].freq_coarse = 2;
		synth_handle->patch.osc[i].freq_fine = 0;
		synth_handle->patch.osc[i].freq_fixed = 0x9449; // pacc nbr (default: A440Hz)
		synth_handle->patch.osc[i].break_pt = 64; // middle of kb
		synth_handle->patch.osc[i].curve_left = 0;
		synth_handle->patch.osc[i].curve_right = 0;
		synth_handle->patch.osc[i].depth_left = 0;
		synth_handle->patch.osc[i].depth_right = 0;
		synth_handle->patch.osc[i].rate_scaling = 0;
		synth_handle->patch.osc[i].output_lvl = 0;
		synth_handle->patch.osc[i].kvs = 0;
		synth_handle->patch.osc[i].enabled = 1;
		synth_handle->patch.osc[i].detune = 7;
		for(j = 0; j < 3; ++j)
		{
			synth_handle->patch.osc[i].eg_rate[j] = 0x0; //encoding for immediate
			synth_handle->patch.osc[i].eg_lvl[j] = 0xFFF;
		}
		// L4 should be 0 in the standard envelope, otherwise key would not stop sounding after release!
		synth_handle->patch.osc[i].eg_lvl[3] = 0x000;
	}
	//set op 1 to max level to always get a basice sine sound as init value
	synth_handle->patch.osc[5].output_lvl = 0xFFF;
	
	//send these parameters to HW
	synth_refresh_hw(synth_handle);
}

void synth_reload(synth_t* synth_handle){

}

void synth_noteoff(synth_t* synth_handle, u8 notevalue, u8 velocity){
	u32 voice;
	for(voice = synth_handle->base_ch; voice < (synth_handle->base_ch + synth_handle->no_ch_avail); ++voice)
	{
		if(synth_handle->ch_array[voice] == (notevalue | 0x80))
		{
			//set key off
			synth_handle->ch_array[voice] &= 0x7F;
			//release if no sustain
			if(!synth_handle->sustain)
				hw_send_others(voice,synth_handle->patch.lfo_speed,synth_handle->patch.lfo_sync,synth_handle->patch.lfo_wave,0);
		}
	}
}


void synth_noteon(synth_t* synth_handle, u8 notevalue, u8 velocity){
	u32 pitch;
	u8 voice;
	u8 i,j;
	u16 lvlresult;
	
	if(!velocity) //workaround for Shadi's buggy midi keyboard
	{
		synth_noteoff(synth_handle, notevalue, velocity);
		return;
	}
	voice = synth_handle->next_ch++;
	//set key_on to off. Since the voice could be occupied already due to low polyphony, 
	//we unset and set the "key on" to retrigger the envelope
	hw_send_others(voice,synth_handle->patch.lfo_speed,synth_handle->patch.lfo_sync,synth_handle->patch.lfo_wave,0);
	if(synth_handle->next_ch == synth_handle->base_ch + synth_handle->no_ch_avail)
		synth_handle->next_ch = synth_handle->base_ch;
	//store that key is played in ch_array. Leftmost bit stores key_on
	synth_handle->ch_array[voice] = notevalue | 0x80;
	synth_send_pitch(synth_handle, voice);
	
	for(i = 0; i < 6; ++i)
	{
		lvlresult = calc_kvs(velocity, synth_handle->patch.osc[i].kvs, synth_handle->patch.osc[i].output_lvl);
		hw_send_op_levelpars(voice, i, lvlresult, synth_handle->patch.osc[i].sync, synth_handle->patch.osc[i].enabled);
		//scale and send envelopes
		for(j = 0; j < 4; ++j)
			hw_send_op_rate(voice, i, j, calc_op_krs(notevalue, synth_handle->patch.osc[i].eg_rate[j], synth_handle->patch.osc[i].rate_scaling));
	}
	//key on
	hw_send_others(voice,synth_handle->patch.lfo_speed,synth_handle->patch.lfo_sync,synth_handle->patch.lfo_wave,1);
}

void synth_cc(synth_t* synth_handle, u8 parameter, u8 value)
{
	switch(parameter)
	{
		case 7: // Volume
			global_volume = (value << 1);
			hw_send_mixer(VOLUME, global_volume);
			break;
		case 10: // Pan
			global_pan = (value << 1);
			hw_send_mixer(PAN, global_pan);
		break;
		case 64: // Sustain
			synth_handle->sustain = value;
			if(!value)
				synth_release_sustained(synth_handle);
		default: break;
	}
}

void synth_pw(synth_t* synth_handle, u16 value)
{
	int voice;
	synth_handle->pitch_bend = value;
	for(voice = synth_handle->base_ch; voice < (synth_handle->base_ch + synth_handle->no_ch_avail);++voice)
	{		
		synth_send_pitch(synth_handle, voice);
	}
}

//This stub is intended for implementing own sysex commands to extend the
//synthesizers MIDI capabilities. It could be used to set up multitimbrality
//and other synth-specific parameters. It is not thought as an alternative to the VST
//interface though and it will be used to control the synthesizer rather than the patch
void synth_decodesysex(u16 SysexRcvIndex){

}

//op and/or stage omitted if not an operator nand/nor stage specific parameter
//The implementation is a little more complicated than needed in som cases. This is
//because the user usually want an immediate respond to a setting while holding down a key
void synth_setparameter(synth_t* synth_handle, u8 op, u8 stage, u32 parType, u32 value)
{
	u8 i,j,ch;
	switch(parType)
	{
		case PAR_AMS:
			synth_handle->patch.osc[op].ams = value;  
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_op_amodsense(ch,op,value);
			}
			break;
		case PAR_MODE:
			synth_handle->patch.osc[op].mode = value;
			break;
		case PAR_SYNC:
			synth_handle->patch.osc[op].sync = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				hw_send_op_levelpars(ch, op, synth_handle->patch.osc[op].output_lvl, value, synth_handle->patch.osc[op].enabled);
			}
			break;
		case PAR_COARSE:
			synth_handle->patch.osc[op].freq_coarse = value;
			break;
		case PAR_FINE:
			synth_handle->patch.osc[op].freq_fine = value;
			break;
		case PAR_FIXED:
			synth_handle->patch.osc[op].freq_fixed = value;
			break;
		case PAR_AEGRATE:
			break;
			synth_handle->patch.osc[op].eg_rate[stage] = value; 
			//rate scaling will set this
			break;
		case PAR_AEGLEVEL:
			synth_handle->patch.osc[op].eg_lvl[stage] = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_op_eglvl(ch, op, stage, value);
			}
			break;
		case PAR_BREAKPT:break;
		case PAR_CL:break;
		case PAR_CR:break;
		case PAR_DL:break;
		case PAR_DR:break;
		case PAR_RS:
			synth_handle->patch.osc[op].rate_scaling = value;
			//SW parameter
			break;
		case PAR_OLEVEL:
			synth_handle->patch.osc[op].output_lvl = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_op_levelpars(ch, op, value, synth_handle->patch.osc[op].sync, synth_handle->patch.osc[op].enabled);
			}
			break;
		case PAR_KVS:
			//SW parameter
			synth_handle->patch.osc[op].kvs = value;
			break;
		case PAR_ENABLED:
			synth_handle->patch.osc[op].enabled = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_op_levelpars(ch, op, value, synth_handle->patch.osc[op].sync, synth_handle->patch.osc[op].enabled);
			}
			break; 
		case PAR_ALG:
			synth_handle->patch.alg = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_algfb(ch,synth_handle->patch.alg,synth_handle->patch.fb);
			}
			break;
		case PAR_FB:break;
			synth_handle->patch.fb = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_algfb(ch,synth_handle->patch.alg,synth_handle->patch.fb);
			}
			break;
		case PAR_LFOWAVE:
			synth_handle->patch.lfo_wave = value;
			// next keypress will send lfo wave. We cannot send anything
			// to hardware here since it could interfere with the playing
			// in terms of key_on and off. This is a result of the parameter
			// grouping
			break;	
		case PAR_LFOSPEED:
			synth_handle->patch.lfo_speed = value;
			//see comment above
			break;
		case PAR_LFODELAY:
			synth_handle->patch.lfo_delay = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_lfo_delay(ch, value);
			}
			break;
		case PAR_LFOSYNC:
			synth_handle->patch.lfo_sync = value;
			//see comment for lfowave
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_lfo_delay(ch, value);
			}
			break;
		case PAR_PMS:
			synth_handle->patch.pms = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_pms(ch, value);
			}
			break;
		case PAR_PEGRATE:
			synth_handle->patch.pitch_eg_rate[stage] = value;
			//Scaling will refresh this
			break;
		case PAR_PEGLEVEL:
			synth_handle->patch.pitch_eg_lvl[stage] = value;
			for(i = 0; i < synth_handle->no_ch_avail; ++i)
			{
				ch = synth_handle->base_ch + i;
				hw_send_peg_lvl(ch, stage, value);
			}
			break;
		case PAR_TRANSPOSE:
			synth_handle->patch.key_transpose = value;
			//SW Parameter
			break;
	}
}

void synth_sendvolume(u8 value)
{
	hw_send_mixer(VOLUME, value);
}

void synth_sendpan(u8 value)
{
	hw_send_mixer(PAN, value);
}
