#include "synth_math.h"
#include "synth.h"


#define PWCENTER (8192)
#define POSBENDCONST (8026)
#define NEGBENDCONST (7150)

// Frequency values
/*
C   8372.0180896192
C#  8869.8441912599
D   9397.2725733570
D#  9956.0634791066
E   10548.0818212118
F   11175.3034058561
F#  11839.8215267723
G   12543.8539514160
G#  13289.7503225582
A   14080.0000000000
A#  14917.2403685788
B   15804.2656401960
*/

// To calculate phase acc increments, use 1048576 
// (phase acc max value, 2^20 since it's 20 bits wide),
// multiply by desired frequency, then multiply by
// (mclk/lrclk) and then divide all by mclk.

// For Fs = 58083.33Hz (mclk=40MHz, mclk/lrclk=768)
 u32 note_map_freq_inc(u8 note){
	static u32 freq_tbl[12] = {
	/* C */ 168551,
	/* C# */ 178574,
	/* D */ 189192,
	/* D# */ 200442,
	/* E */ 212361,
	/* F */ 224989,
	/* F# */ 238367, 
	/* G */ 252541,
	/* G# */ 267558,
	/* A */ 283467,
	/* A# */ 300324,
	/* B */ 318182
	};
	return (freq_tbl[note%12]) >> ((131-note)/12);
}

u32 calc_freq(u8 note, u8 transpose, u8 ratio, u8 fine)
{
	return ((note_map_freq_inc((note + TRANSPOSE_MID) - transpose) >> 1 ) * ratio * (100 + fine))/100;
}

//Since bending is implemented in the fixed range of 2 semitones, it could
//be calculated linearly without too much disturbance. If we want to allow
//greater ranges, exponential math is required.
u32 apply_pw(u32 basefreq, u16 wheelval)
{
	u32 bendamt;
	u32 bendfreqamt;
	if(wheelval >= PWCENTER)
	{
		bendamt = ((wheelval - PWCENTER) * POSBENDCONST) >> 15;//14 + 16 - 16 = 14 bits for bending
		bendfreqamt = (bendamt * basefreq) >> 14;
		return basefreq + bendfreqamt;
	} else {
		bendamt = ((PWCENTER - wheelval) * NEGBENDCONST) >> 15;
		bendfreqamt = (bendamt * basefreq) >> 14;
		return basefreq - bendfreqamt;
	}
}

//This function linearly interpolates the veocity sensitivty in a hard-coded manner.
//It is derived from the tables in the hexter synth, but calculated as a function in
//order to keep memory usage low
u16 calc_kvs(u8 vel, u32 velsense, u32 outlevel)
{
	u32 reduction;
	reduction = (((127 - vel) * velsense * outlevel)/127) >> 13;
	 // I've divided into 3 linear segments, first 128 down to 64 will reduce up to 25%,
	 // 64 down to 24 will reduce up to 50% and then the third segments will reduce to 100
	if(vel < 64)
		reduction = (((172 - (vel << 1)) * velsense * outlevel)/172) >> 12;
	else if(vel < 24)
		reduction = (((176 - (vel << 2)) * velsense * outlevel)/176) >> 12;
	return (outlevel - reduction);
}

//rate scaling is calculated up to note 102. At maximum, note 102 and above gives +42 to rate value
//The table used for rate multipliers is an average of the ratio of going 42 indexes forward in the
//rate tables in dx7_converter.c
u32 calc_op_krs(u8 note, u32 baserate, u8 scaleamount)
{
	u32 rval;
	//8.8 format. Integer and frac parts separated
	static u8 op_rsimults[8] = 
	{
		0,1,3,7,14,27,49,109
	};

	static u8 op_rsfmults[8] = 
	{
		0xD8,0x6D,0x77,0x1C,0x64,0xBB,0x06
	};
	u32 extradec, extrafrac;
	if(note > 102)
		note = 102;
	extradec = ((op_rsimults[scaleamount] * note) / 102) * baserate;
	extrafrac = ((op_rsfmults[scaleamount] * note) / 102) * baserate;
	rval = baserate + extradec + (extrafrac >> 8);
	return rval > 101182052 ? 101182052 : rval; // env rate cannot be grater than dx7 "99"
}

